function epnose()
%   EPNOSE
%   Evaluation of Phase in uNcertain and nOisy SEries
%   Co. B.Ruzek & J.Laurin, July 2017
%   Version 1.0
%
%   Please cite as:
%   Laurin, J., Ruzek, B., Giorgioni, M., 2017. Orbital signals in carbon isotopes:	
%   phase distortion as a signature of the carbon cycle. Paleoceanography.
% ------------------------------------

%pkg load signal; % This line should be uncommented for running in Octave

    function [Cfg] = init_fcn()
        
        function [Cfg] = check_input(Cfg)
            chk_flg = 0;
            if(Cfg.ntrial < 0)
                Cfg.ntrial = 0;
                fprintf(1, 'Warning: Number of Monte Carlo simulations set to 0\n');
            end
            if(Cfg.Fcen_a <= 0 || Cfg.Fcen_b <= 0)
                chk_flg = chk_flg + 1;
                fprintf(1, 'Central frequency must be strictly positive, exitting....\n');
            end
            if(Cfg.Fwid_a >= Cfg.Fcen_a)
                chk_flg = chk_flg + 1;
                fprintf(1, 'Filter half-width must be higher than its central frequency, exitting....\n');
            end
            if(Cfg.Fwid_b >= Cfg.Fcen_b)
                chk_flg = chk_flg + 1;
                fprintf(1, 'Filter half-width must be higher than its central frequency, exitting....\n');
            end
            if(Cfg.t_err < 0 || Cfg.x_err < 0)
                chk_flg = chk_flg + 1;
                fprintf(1, 'Both t and x errors must be non-negative, exitting....\n');
            end
            if(Cfg.FltId < 1 || Cfg.FltId > 3)
                chk_flg = chk_flg + 1;
                fprintf(1, 'Filter id must 1|2|3, exitting...\n');
            end
            if(chk_flg ~= 0)
                Cfg = [];
            end
        end
        
        close('all');
        if(exist('epnose.mat', 'file'))
            load('epnose.mat');
        else       
            Cfg.input_file = 'tables/SIMB_N1a.txt'; % file to process
            Cfg.ntrial = 1;   % number of Monte-Carlo numerical simulations
            Cfg.Fcen_a = 9.1;  % central frequency of filter
            Cfg.Fwid_a = 3.0;  % half-width of filter
            Cfg.Fcen_b = 2.47;  % central frequency of filter
            Cfg.Fwid_b = 0.50;  % half-width of filter
            Cfg.t_err  = 0;  % relative t-axis error
            Cfg.x_err  = 0;  % relative x-axis error
            Cfg.FltId  = 3;  % specification of the filter used
        end
        
        prompt = {...
            'File name with time-series data:',...
            'Number of Monte carlo samples:'...
            'Central frequency of 1st filter (envelope)', ...
            'Half-width of 1st filter',...
            'Central frequency of 2nd filter (plain signal)', ...
            'Half-width of 1st filter',...
            'X-axis uncertainty (0-1)*dx',...
            'Y-axis uncertainty (units of Y)',...
            'Filter specification (1=Gauss,2=Taner orig,3=Taner linear)',...
            };
        dlg_title = 'Input';
        num_lines = 1;
        defaultans = {...
            sprintf('%s',Cfg.input_file),...
            sprintf('%d',Cfg.ntrial),...
            sprintf('%f',Cfg.Fcen_a),...
            sprintf('%f',Cfg.Fwid_a),...
            sprintf('%f',Cfg.Fcen_b),...
            sprintf('%f',Cfg.Fwid_b),...
            sprintf('%f',Cfg.t_err),...
            sprintf('%f',Cfg.x_err),...
            sprintf('%d',Cfg.FltId),...
            };
        answer = inputdlg(prompt,dlg_title,num_lines,defaultans);
        if(~isempty(answer))
            Cfg.input_file = char(answer{1});
            Cfg.ntrial = str2double(answer{2});
            Cfg.Fcen_a = str2double(answer{3});
            Cfg.Fwid_a = str2double(answer{4});
            Cfg.Fcen_b = str2double(answer{5});
            Cfg.Fwid_b = str2double(answer{6});
            Cfg.t_err  = str2double(answer{7});
            Cfg.x_err  = str2double(answer{8});
            Cfg.FltId  = str2double(answer{9});
            Cfg = check_input(Cfg);
            if(isempty(Cfg))
                return
            end
            save('epnose.mat', 'Cfg', '-v7');
        else
            Cfg = [];
            return;
        end
        fprintf(1, '=================================================\n');
        fprintf(1, 'Data table:  %s\n', Cfg.input_file);
        fprintf(1, 'Number of Monte Carlo samples  %d\n', Cfg.ntrial);
        fprintf(1, 'Filter #1: central frequency %6.3f half_width %6.3f\n', Cfg.Fcen_a, Cfg.Fwid_a);
        fprintf(1, 'Filter #2: central frequency %6.3f half_width %6.3f\n', Cfg.Fcen_b, Cfg.Fwid_b);
        fprintf(1, '=================================================\n');
        fprintf(fpo, '=================================================\n');
        fprintf(fpo, 'Data table:  %s\n', Cfg.input_file);
        fprintf(fpo, 'Number of Monte Carlo samples  %d\n', Cfg.ntrial);
        fprintf(fpo, 'Filter #1: central frequency %6.3f half_width %6.3f\n', Cfg.Fcen_a, Cfg.Fwid_a);
        fprintf(fpo, 'Filter #2: central frequency %6.3f half_width %6.3f\n', Cfg.Fcen_b, Cfg.Fwid_b);
        fprintf(fpo, '=================================================\n');
        
    end

    function [t,x] = read_table(input_file)
        t = [];
        x = [];
        try
            [t,x] = textread(input_file, '%f %f');
            %[t,x] = textscan(input_file, '%f %f');
        catch
            fprintf(1, 'Input file %s not accessible for reading\n', input_file);
            fprintf(1, 'Exitting ...\n');
            return;
        end
        indOK = find(~isnan(t) & ~isnan(x));
        t = t(indOK);
        x = x(indOK);
        x = x - mean(x);
        fprintf(1, 'Time span      %f - %f\n', min(t), max(t));
        fprintf(1, 'Function range %f - %f\n', min(x), max(x));
        fprintf(1, 'Number of samples = %d | %d\n', length(t), length(x));
        fprintf(1, 'Average sampling interval = %f\n', (max(t)-min(t))/(length(t)-1));
        fprintf(1, '=================================================\n');
        fprintf(fpo, 'Time span      %f - %f\n', min(t), max(t));
        fprintf(fpo, 'Function range %f - %f\n', min(x), max(x));
        fprintf(fpo, 'Number of samples = %d\n', length(t));
        fprintf(fpo, 'Average sampling interval = %f\n', (max(t)-min(t))/(length(t)-1));
        fprintf(fpo, '=================================================\n');
        
    end

    function [to,xo] = resample(ti,xi)
        % resampling approach:
        % minimum and maximum times are preserved
        nsamp = length(ti);
        minT  = min(ti);
        maxT  = max(ti);
        delT  = (maxT - minT)/(nsamp-1);
        to = minT:delT:maxT;
        xo = interp1(ti,xi,to);
    end

    function [to,xo] = add_noise(ti,xi)
        % noise is added both to time and signal
        % NOTE : time need not to be monotonous function after applying the
        % noise. Resampling may solve this issue
        nsamp = length(ti);
        dt = mean(diff(ti));
        %        dx = std(xi);
        to = ti + Cfg.t_err*(rand(nsamp,1)-0.5)*dt;
        xo = xi + Cfg.x_err*(rand(nsamp,1)-0.5);
        [~,ind] = sort(to);
        to = to(ind);
        xo = xo(ind);
    end

    function [t_out, x_out] = G_filt(wid,label,t_in, x_in, f_cen, f_wid)
        % Common plain Gaussian filter
        nt      = length(t_in);                       % number of times
        nx      = length(x_in);                       % number of samples
        if(nt ~= nx)
            % Both vectors must have the same size
            fprintf(1, 'Inproper number of time/function samples, exitting ...\n');
            exit(1);
        end
        t2t = diff(diff(t_in));
        dt = t_in(2) - t_in(1);
        if((max(t2t) - min(t2t)) > dt*1.e-6)
            % t_in must represent regular sampling
            fprintf(1, 'Irregular time sampling, exitting ...\n');
            exit(1);
        end
        fs    = 1/dt;                               % sampling frequency
        df    = fs/nt;                              % frequency step
        X_in  = fft(x_in);                          % signal spectrum
        ny    = floor(length(X_in)/2) + 1;          % number of 'positive' frequencies
        f     = (0:ny-1)*df;                        % set of positive frequencies
        f_all = (0:nx-1)*df;
        f_all(ny:end) = f_all(ny:end) - fs;         % all frequencied required
        %f_wid = 1/(T_cen - T_wid) - 1/(T_cen + T_wid);
        H     = exp(-((f-f_cen)/f_wid).^2);
        H_all = exp(-((abs(f_all)-f_cen)/f_wid).^2);
        t_out = t_in;
        X_out = X_in.*H_all;
        x_out = real(ifft(X_out));
        if(~isempty(label))
            plot_spec(wid,label,f,X_in,H,X_out);
        end
    end

    function [t_out, x_out] = L_filt(wid,label,t_in, x_in, f_cen, f_wid)
        % Modified Taner filter which is symmetric around f_cen.
        %    function [t_out, x_out] = filtX(wid,label,t_in, x_in, f_cen, f_wid)
        nt      = length(t_in);                       % number of times
        nx      = length(x_in);                       % number of samples
        if(nt ~= nx)
            % Both vectors must have the same size
            fprintf(1, 'Inproper number of time/function samples, exitting ...\n');
            exit(1);
        end
        t2t = diff(diff(t_in));
        dt = t_in(2) - t_in(1);
        if((max(t2t) - min(t2t)) > dt*1.e-6)
            % t_in must represent regular sampling
            fprintf(1, 'Irregular time sampling, exitting ...\n');
            exit(1);
        end
        fs    = 1/dt;                               % sampling frequency
        df    = fs/nt;                              % frequency step
        X_in  = fft(x_in);                          % signal spectrum
        ny    = floor(length(X_in)/2) + 1;          % number of 'positive' frequencies
        f     = (0:ny-1)*df;                        % set of positive frequencies
        f_all = (0:nx-1)*df;
        f_all(ny:end) = f_all(ny:end) - fs;         % all frequencies required
        dbh = 48;
        bh  = log(0.5)/2;
        ah  = log((-3.0103-dbh/10)*log(10)/bh);
        H = exp(bh*(abs(f-f_cen)/f_wid).^(ah/log(2)));
        H_all = exp(bh*(abs(f_all-f_cen)/f_wid).^(ah/log(2)));
        H   = H/max(H);
        H_all = H_all/max(H_all);
        t_out = t_in;
        X_out = X_in.*H_all;
        x_out = real(ifft(X_out));
        if(~isempty(label))
            plot_spec(wid,label,f,X_in,H,X_out);
        end
    end

    function [t_out, x_out] = T_filt(wid,label,t_in, x_in, f_cen, f_wid)
        % Taner filter as originally formulated, i.e.it is symmetric aroung f_cen only
        % in logaritmnic frequency scaling
        nt      = length(t_in);                       % number of times
        nx      = length(x_in);                       % number of samples
        if(nt ~= nx)
            % Both vectors must have the same size
            fprintf(1, 'Inproper number of time/function samples, exitting ...\n');
            exit(1);
        end
        t2t = diff(diff(t_in));
        dt = t_in(2) - t_in(1);
        if((max(t2t) - min(t2t)) > dt*1.e-6)
            % t_in must represent regular sampling
            fprintf(1, 'Irregular time sampling, exitting ...\n');
            exit(1);
        end
        fs    = 1/dt;                               % sampling frequency
        df    = fs/nt;                              % frequency step
        X_in  = fft(x_in);                          % signal spectrum
        ny    = floor(length(X_in)/2) + 1;          % number of 'positive' frequencies
        f     = (0:ny-1)*df;                        % set of positive frequencies
        f_all = (0:nx-1)*df;
        f_all(ny:end) = f_all(ny:end) - fs;         % all frequencied required
        
        xf    = log2(f(2:end));
        x_all = log2(abs(f_all(2:end)));
        
        dbh = 48;
        xl  = log2(f_cen + f_wid);
        bl  = log(0.5)/2;
        al  = log((-3.0103-dbh/10)*log(10)/bl);
        xh  = log2(f_cen - f_wid);
        bh  = log(0.5)/2;
        ah  = log((-3.0103-dbh/10)*log(10)/bh);
        H   = [0 exp(bl*exp(al*(xf-xl)) + bh*exp(ah*(xh-xf)))];
        H_all = [0 exp(bl*exp(al*(x_all-xl)) + bh*exp(ah*(xh-x_all)))];
        H   = H/max(H);
        H_all = H_all/max(H_all);
        t_out = t_in;
        X_out = X_in.*H_all;
        x_out = real(ifft(X_out));
        if(~isempty(label))
            plot_spec(wid,label,f,X_in,H,X_out);
        end
    end

    function [x_lst,y_lst,f_lst] = peak_lst(x,y,n)
        yy = y;
        x_lst = [];
        y_lst = [];
        f_lst = [];
        n_lst = 0;
        i_max = length(y);
        while(n_lst < n)
            ymax = max(y);
            ii = find(y == ymax);
            if(isempty(ii))
                break;
            end
            ii = ii(1);
            x_lst = [x_lst; x(ii)];
            y_lst = [y_lst; y(ii)];
            f_lst = [f_lst; +1];
            n_lst = n_lst + 1;
            i_l = ii-1;
            while(i_l > 1)
                if(y(i_l) < y(i_l-1))
                    break;
                end
                y(i_l) = nan;
                i_l = i_l - 1;
            end
            i_r = ii + 1;
            while(i_r < i_max)
                if(y(i_r) < y(i_r+1))
                    break;
                end
                y(i_r) = nan;
                i_r = i_r + 1;
            end
            y(ii) = nan;
        end
        y = -yy;
        while(n_lst < n)
            ymax = max(y);
            ii = find(y == ymax);
            if(isempty(ii))
                break;
            end
            ii = ii(1);
            x_lst = [x_lst; x(ii)];
            y_lst = [y_lst;-y(ii)];
            f_lst = [f_lst;-1];
            n_lst = n_lst + 1;
            i_l = ii-1;
            while(i_l > 1)
                if(y(i_l) < y(i_l-1))
                    break;
                end
                y(i_l) = nan;
                i_l = i_l - 1;
            end
            i_r = ii + 1;
            while(i_r < i_max)
                if(y(i_r) < y(i_r+1))
                    break;
                end
                y(i_r) = nan;
                i_r = i_r + 1;
            end
            y(ii) = nan;
        end
        indOK = find(x_lst > min(x) & x_lst < max(x));
        x_lst = x_lst(indOK);
        y_lst = y_lst(indOK);
        f_lst = f_lst(indOK);
        [x_lst, x_ind] = sort(x_lst);
        y_lst = y_lst(x_ind);
        f_lst = f_lst(x_ind);
    end

    function plot_spec(wid, label,f,X_in,H,X_out)
        hdl = figure(1);
        set(hdl, 'Position', [0,300,500,700]);
        if(wid == 1)
            wid4 = 1;
        else
            wid4 = 2;
        end
        subplot(2,1,wid4);
        hold 'on';
        nf = length(f);
        A_in  = abs(X_in(1:nf));
        A_out = abs(X_out(1:nf));
        A_max = max([max(A_in) max(A_in)]);
        plot(f, A_in/A_max, 'r');
        plot(f, A_out/A_max, 'b');
        H_rsp = abs(H(1:nf));
        H_rsp = H_rsp/max(H_rsp);
        plot(f, H_rsp, 'k');
        %fpx = fopen('spec.dat','at');
        %for iii=1:length(f)
        %fprintf(fpx, '%f %f\n', f(iii), H_rsp(iii));
        %end
        %fclose(fpx);
        
        title(label);
        xlabel('Frequency');
        grid 'on';
    end

    function plot_signal(wid, label, tr,xr,tf,xf)
        hdl = figure(wid);
        if(wid == 2)
            set(hdl, 'Position', [500,500,500,500]);
        end
        if(wid == 3)
            set(hdl, 'Position', [500,0,500,500]);
        end
        
        subplot(2,1,1);
        hold 'on';
        plot(tr, xr, 'r');
        plot(tf, xf, 'b');
        title(label);
        xlabel('X (Time)');
        ylabel('Y');
        grid 'on';
    end

    function plot_peaks(wid, label, t,x,tpos,xpos,flg)
        figure(wid);
        subplot(2,1,2);
        hold 'on';
        plot(t, x, 'b');
        for ii=1:length(xpos)
            gh = text(tpos(ii),xpos(ii), sprintf('%d',ii),'HorizontalAlignment','center','VerticalAlignment','middle');
            if(flg(ii) > 0)
                set(gh,'color', [1 0 0])
            end
            if(flg(ii) < 0)
                set(gh,'color', [0 1 0])
            end
        end
        title(label);
        xlabel('X (Time)');
        ylabel('Y');
        grid 'on';
    end

    function [x_histA, x_histB] = collect_picks(t_,x_,xposA,yposA,flgA,xposB,yposB,flgB)
        [tr0,xr0]   = resample(t_,x_);
        [tfA0,xfA0] = filtX(1,'',tr0,xr0,Cfg.Fcen_a,Cfg.Fwid_a);
        [tfB0,xfB0] = filtX(2,'', tr0,xr0,Cfg.Fcen_b,Cfg.Fwid_b);
        sig_env0   = abs(hilbert(xfA0));
        x_histA = [];
        x_histB = [];
        n_class = 20*length(t_);
        for ii=1:Cfg.ntrial
            [tn,xn]   = add_noise(t_,x_);
            [tr,xr]   = resample(tn,xn);
            [tfA,xfA] = filtX(1,'',tr,xr,Cfg.Fcen_a,Cfg.Fwid_a);        % 1-st filtered version of [tr,xr]
            [tfB,xfB] = filtX(2,'',tr,xr,Cfg.Fcen_b,Cfg.Fwid_b);        % 2-nd filtered version of [tr,xr]
            sig_env   = abs(hilbert(xfA));
            [x_posAx,~] = peak_lst(tfA,sig_env,2000);
            [x_posBx,~] = peak_lst(tfB,xfB,2000);
            x_histA = [x_histA; x_posAx];
            x_histB = [x_histB; x_posBx];
        end
        x_histA = sort(x_histA);
        x_histB = sort(x_histB);
        
        hdl = figure(4);
        set(hdl, 'Position', [1000 300 700 700]);
        subplot(4,1,1);
        hold 'on';
        hist(x_histA, n_class);
        title('1st signal picks');
        grid 'on';
        subplot(4,1,2);
        hold 'on';
        plot(tfA0, sig_env0);
        title('1st signal');
    
        for ii=1:length(xposA)
            gh = text(xposA(ii),yposA(ii), sprintf('%d',ii),'HorizontalAlignment','center','VerticalAlignment','middle');
            if(flgA(ii) > 0)
                set(gh,'color', [1 0 0])
            end
            if(flgA(ii) < 0)
                set(gh,'color', [0 1 0])
            end
        end
        
        grid 'on';
        subplot(4,1,3);
        hold 'on';
        hist(x_histB, n_class);
        title('2nd signal picks');
        grid 'on';
        subplot(4,1,4);
        hold 'on';
        plot(tfB0, xfB0);
        title('2nd signal');
        
        for ii=1:length(xposB)
            gh = text(xposB(ii),yposB(ii), sprintf('%d',ii),'HorizontalAlignment','center','VerticalAlignment','middle');
            if(flgB(ii) > 0)
                set(gh,'color', [1 0 0])
            end
            if(flgB(ii) < 0)
                set(gh,'color', [0 1 0])
            end
        end
        
        grid 'on';
    end

    function analyse_picks(title, master_picks, flg, trial_picks)
        fprintf(1, '%s\n', title);
        fprintf(1,' Nr  Pol   X(Time) MC_mean   MC_std\n');
        fprintf(fpo, '%s\n', title);
        fprintf(fpo,' Nr  Pol   X(Time) MC_mean   MC_std\n');
        master_picks = sort(master_picks);
        n_master = length(master_picks);
        for ii=1:n_master
            dist    = trial_picks - master_picks(ii);
            distA   = abs(dist);
            [~,ind] = sort(distA,'ascend');
            distx   = dist(ind(1:Cfg.ntrial));
            pick_m  = mean(distx) + master_picks(ii);
            pick_s  = std(distx);
            fprintf(1, '%3d (%+d) %8.3f %8.3f %8.3f\n', ii, flg(ii), master_picks(ii), pick_m, pick_s);
            fprintf(fpo, '%3d (%+d) %8.3f %8.3f %8.3f\n', ii, flg(ii), master_picks(ii), pick_m, pick_s);
        end
    end

    function save_fsig(tA,xA, tB,xB)
        fpl = fopen('sig1.dat', 'wt');
        for iis=1:length(tA)
            fprintf(fpl, '%f %f\n', tA(iis), xA(iis));
        end
        fclose(fpl);
        fpl = fopen('sig2.dat', 'wt');
        for iis=1:length(tB)
            fprintf(fpl, '%f %f\n', tB(iis), xB(iis));
        end
        fclose(fpl);
        
    end

% ****************************************************
% *********** End of function declarations ***********
% ****************************************************

fpo = fopen('epnose.log', 'wt');
Cfg       = init_fcn();
if(isempty(Cfg))
    return;
end
switch(Cfg.FltId)
    case 1
        filtX = @G_filt;
    case 2
        filtX = @T_filt;
    case 3
        filtX = @L_filt;
end
[t,x]     = read_table(Cfg.input_file);
[tr,xr]   = resample(t,x);                      % time series interpolated to regular sampling
[tfA,xfA] = filtX(1,'1-st filter',tr,xr,Cfg.Fcen_a,Cfg.Fwid_a);        % 1-st filtered version of [tr,xr]
[tfB,xfB] = filtX(2,'2nd-filter', tr,xr,Cfg.Fcen_b,Cfg.Fwid_b);        % 2-nd filtered version of [tr,xr]
sig_env   = abs(hilbert(xfA));

plot_signal(2,'1st signal', tr,xr,tfA,xfA);
[x_posA,y_posA,flgA] = peak_lst(tfA,sig_env,2000);
plot_peaks(2,'Peaks on filtered signal envelope', tfA,sig_env,x_posA,y_posA,flgA);

plot_signal(3,'2nd signal', tr,xr,tfB,xfB);
[x_posB,y_posB,flgB] = peak_lst(tfB,xfB,2000);
plot_peaks(3,'Peaks on filtered signal', tfB,xfB,x_posB,y_posB,flgB);

[x_posa, x_posb] = collect_picks(t,x,x_posA,y_posA,flgA,x_posB,y_posB,flgB);
analyse_picks('1st time series', x_posA, flgA, x_posa);
analyse_picks('2nd time series', x_posB, flgB, x_posb);

fclose(fpo);

save_fsig(tfA,sig_env, tfB,xfB);

end
